@@ -140,19 +140,19 @@ pub(crate) fn datetime_repr(d: &DateTime<Utc>) -> String {
 #[derive(Clone, Debug, Serialize, Deserialize)]
 struct PythonCredentialsFetcher<CredType> {
     pub pickled_function: Vec<u8>,
-    pub current: Option<CredType>,
+    pub initial: Option<CredType>,
 }
 
 impl<CredType> PythonCredentialsFetcher<CredType> {
     fn new(pickled_function: Vec<u8>) -> Self {
-        PythonCredentialsFetcher { pickled_function, current: None }
+        PythonCredentialsFetcher { pickled_function, initial: None }
     }
 
-    fn new_with_current<C>(pickled_function: Vec<u8>, current: C) -> Self
+    fn new_with_initial<C>(pickled_function: Vec<u8>, current: C) -> Self
     where
         C: Into<CredType>,
     {
-        PythonCredentialsFetcher { pickled_function, current: Some(current.into()) }
+        PythonCredentialsFetcher { pickled_function, initial: Some(current.into()) }
     }
 }
 
@@ -174,18 +174,22 @@ where
 #[typetag::serde]
 impl S3CredentialsFetcher for PythonCredentialsFetcher<S3StaticCredentials> {
     async fn get(&self) -> Result<S3StaticCredentials, String> {
-        Python::with_gil(|py| {
-            if let Some(static_creds) = self.current.as_ref() {
-                // avoid herding by adding some randomness
-                let delta = TimeDelta::seconds(rand::random_range(5..180));
-                let expiration = static_creds
-                    .expires_after
-                    .map(|exp| exp - delta)
-                    .unwrap_or(chrono::DateTime::<Utc>::MAX_UTC);
-                if Utc::now() < expiration {
+        if let Some(static_creds) = self.initial.as_ref() {
+            match static_creds.expires_after {
+                None => {
                     return Ok(static_creds.clone());
                 }
+                Some(expiration)
+                    if expiration
+                        > Utc::now()
+                            + TimeDelta::seconds(rand::random_range(120..=180)) =>
+                {
+                    return Ok(static_creds.clone());
+                }
+                _ => {}
             }
+        }
+        Python::with_gil(|py| {
             call_pickled::<PyS3StaticCredentials>(py, self.pickled_function.clone())
                 .map(|c| c.into())
         })
@@ -197,18 +201,22 @@ impl S3CredentialsFetcher for PythonCredentialsFetcher<S3StaticCredentials> {
 #[typetag::serde]
 impl GcsCredentialsFetcher for PythonCredentialsFetcher<GcsBearerCredential> {
     async fn get(&self) -> Result<GcsBearerCredential, String> {
-        Python::with_gil(|py| {
-            if let Some(static_creds) = self.current.as_ref() {
-                // avoid herding by adding some randomness
-                let delta = TimeDelta::seconds(rand::random_range(5..180));
-                let expiration = static_creds
-                    .expires_after
-                    .map(|exp| exp - delta)
-                    .unwrap_or(chrono::DateTime::<Utc>::MAX_UTC);
-                if Utc::now() < expiration {
+        if let Some(static_creds) = self.initial.as_ref() {
+            match static_creds.expires_after {
+                None => {
                     return Ok(static_creds.clone());
                 }
+                Some(expiration)
+                    if expiration
+                        > Utc::now()
+                            + TimeDelta::seconds(rand::random_range(120..=180)) =>
+                {
+                    return Ok(static_creds.clone());
+                }
+                _ => {}
             }
+        }
+        Python::with_gil(|py| {
             call_pickled::<PyGcsBearerCredential>(py, self.pickled_function.clone())
                 .map(|c| c.into())
         })
@@ -233,7 +241,7 @@ impl From<PyS3Credentials> for S3Credentials {
             PyS3Credentials::Static(creds) => S3Credentials::Static(creds.into()),
             PyS3Credentials::Refreshable { pickled_function, current } => {
                 let fetcher = if let Some(current) = current {
-                    PythonCredentialsFetcher::new_with_current(pickled_function, current)
+                    PythonCredentialsFetcher::new_with_initial(pickled_function, current)
                 } else {
                     PythonCredentialsFetcher::new(pickled_function)
                 };
@@ -278,7 +286,9 @@ impl From<PyGcsStaticCredentials> for GcsStaticCredentials {
 #[pyclass(name = "GcsBearerCredential")]
 #[derive(Clone, Debug)]
 pub struct PyGcsBearerCredential {
+    #[pyo3(get)]
     pub bearer: String,
+    #[pyo3(get)]
     pub expires_after: Option<DateTime<Utc>>,
 }
 
@@ -306,6 +316,7 @@ impl From<GcsBearerCredential> for PyGcsBearerCredential {
 #[pyclass(name = "GcsCredentials")]
 #[derive(Clone, Debug)]
 pub enum PyGcsCredentials {
+    Anonymous(),
     FromEnv(),
     Static(PyGcsStaticCredentials),
     Refreshable { pickled_function: Vec<u8>, current: Option<PyGcsBearerCredential> },
@@ -314,11 +325,12 @@ pub enum PyGcsCredentials {
 impl From<PyGcsCredentials> for GcsCredentials {
     fn from(value: PyGcsCredentials) -> Self {
         match value {
+            PyGcsCredentials::Anonymous() => GcsCredentials::Anonymous,
             PyGcsCredentials::FromEnv() => GcsCredentials::FromEnv,
             PyGcsCredentials::Static(creds) => GcsCredentials::Static(creds.into()),
             PyGcsCredentials::Refreshable { pickled_function, current } => {
                 let fetcher = if let Some(current) = current {
-                    PythonCredentialsFetcher::new_with_current(pickled_function, current)
+                    PythonCredentialsFetcher::new_with_initial(pickled_function, current)
                 } else {
                     PythonCredentialsFetcher::new(pickled_function)
                 };
@@ -400,31 +412,43 @@ pub struct PyS3Options {
     pub anonymous: bool,
     #[pyo3(get, set)]
     pub force_path_style: bool,
+    #[pyo3(get, set)]
+    pub network_stream_timeout_seconds: Option<u32>,
 }
 
 #[pymethods]
 impl PyS3Options {
     #[new]
-    #[pyo3(signature = ( region=None, endpoint_url=None, allow_http=false, anonymous=false, force_path_style=false))]
+    #[pyo3(signature = ( region=None, endpoint_url=None, allow_http=false, anonymous=false, force_path_style=false, network_stream_timeout_seconds=None))]
     pub(crate) fn new(
         region: Option<String>,
         endpoint_url: Option<String>,
         allow_http: bool,
         anonymous: bool,
         force_path_style: bool,
+        network_stream_timeout_seconds: Option<u32>,
     ) -> Self {
-        Self { region, endpoint_url, allow_http, anonymous, force_path_style }
+        Self {
+            region,
+            endpoint_url,
+            allow_http,
+            anonymous,
+            force_path_style,
+            network_stream_timeout_seconds,
+        }
     }
 
     pub fn __repr__(&self) -> String {
         // TODO: escape
         format!(
-            r#"S3Options(region={region}, endpoint_url={url}, allow_http={http}, anonymous={anon}, force_path_style={force_path_style})"#,
+            r#"S3Options(region={region}, endpoint_url={url}, allow_http={http}, anonymous={anon}, force_path_style={force_path_style}, network_stream_timeout_seconds={net_timeout})"#,
             region = format_option(self.region.as_ref()),
             url = format_option(self.endpoint_url.as_ref()),
             http = format_bool(self.allow_http),
             anon = format_bool(self.anonymous),
             force_path_style = format_bool(self.force_path_style),
+            net_timeout =
+                format_option(self.network_stream_timeout_seconds.map(|n| n.to_string())),
         )
     }
 }
@@ -437,6 +461,7 @@ impl From<&PyS3Options> for S3Options {
             allow_http: options.allow_http,
             anonymous: options.anonymous,
             force_path_style: options.force_path_style,
+            network_stream_timeout_seconds: options.network_stream_timeout_seconds,
         }
     }
 }
@@ -449,6 +474,7 @@ impl From<S3Options> for PyS3Options {
             allow_http: value.allow_http,
             anonymous: value.anonymous,
             force_path_style: value.force_path_style,
+            network_stream_timeout_seconds: value.network_stream_timeout_seconds,
         }
     }
 }
@@ -1419,6 +1445,8 @@ pub struct PyRepositoryConfig {
     #[pyo3(get, set)]
     pub compression: Option<Py<PyCompressionConfig>>,
     #[pyo3(get, set)]
+    pub max_concurrent_requests: Option<u16>,
+    #[pyo3(get, set)]
     pub caching: Option<Py<PyCachingConfig>>,
     #[pyo3(get, set)]
     pub storage: Option<Py<PyStorageSettings>>,
@@ -1454,6 +1482,7 @@ impl TryFrom<&PyRepositoryConfig> for RepositoryConfig {
                 inline_chunk_threshold_bytes: value.inline_chunk_threshold_bytes,
                 get_partial_values_concurrency: value.get_partial_values_concurrency,
                 compression: value.compression.as_ref().map(|c| (&*c.borrow(py)).into()),
+                max_concurrent_requests: value.max_concurrent_requests,
                 caching: value.caching.as_ref().map(|c| (&*c.borrow(py)).into()),
                 storage: value.storage.as_ref().map(|s| (&*s.borrow(py)).into()),
                 virtual_chunk_containers: cont,
@@ -1473,6 +1502,7 @@ impl From<RepositoryConfig> for PyRepositoryConfig {
                 Py::new(py, Into::<PyCompressionConfig>::into(c))
                     .expect("Cannot create instance of CompressionConfig")
             }),
+            max_concurrent_requests: value.max_concurrent_requests,
             caching: value.caching.map(|c| {
                 Py::new(py, Into::<PyCachingConfig>::into(c))
                     .expect("Cannot create instance of CachingConfig")
@@ -1502,12 +1532,13 @@ impl PyRepositoryConfig {
     }
 
     #[new]
-    #[pyo3(signature = (inline_chunk_threshold_bytes = None, get_partial_values_concurrency = None, compression = None, caching = None, storage = None, virtual_chunk_containers = None, manifest = None))]
+    #[pyo3(signature = (inline_chunk_threshold_bytes = None, get_partial_values_concurrency = None, compression = None, max_concurrent_requests = None, caching = None, storage = None, virtual_chunk_containers = None, manifest = None))]
     #[allow(clippy::too_many_arguments)]
     pub fn new(
         inline_chunk_threshold_bytes: Option<u16>,
         get_partial_values_concurrency: Option<u16>,
         compression: Option<Py<PyCompressionConfig>>,
+        max_concurrent_requests: Option<u16>,
         caching: Option<Py<PyCachingConfig>>,
         storage: Option<Py<PyStorageSettings>>,
         virtual_chunk_containers: Option<HashMap<String, PyVirtualChunkContainer>>,
@@ -1517,6 +1548,7 @@ impl PyRepositoryConfig {
             inline_chunk_threshold_bytes,
             get_partial_values_concurrency,
             compression,
+            max_concurrent_requests,
             caching,
             storage,
             virtual_chunk_containers,
