@@ -17,18 +17,19 @@ use icechunk::{
         snapshot::{ManifestFileInfo, SnapshotInfo, SnapshotProperties},
         transaction_log::Diff,
     },
+    inspect::snapshot_json,
     ops::{
         gc::{ExpiredRefAction, GCConfig, GCSummary, expire, garbage_collect},
         manifests::rewrite_manifests,
         stats::repo_chunks_storage,
     },
-    repository::{RepositoryErrorKind, VersionInfo},
+    repository::{RepositoryError, RepositoryErrorKind, VersionInfo},
 };
 use pyo3::{
     IntoPyObjectExt,
     exceptions::PyValueError,
     prelude::*,
-    types::{PyDict, PyNone, PyType},
+    types::{PyDict, PyNone, PySet, PyType},
 };
 use serde::{Deserialize, Serialize};
 use tokio::sync::{Mutex, RwLock};
@@ -74,8 +75,11 @@ impl_pickle!(PySnapshotInfo);
 #[pyclass(name = "ManifestFileInfo", eq)]
 #[derive(Clone, Debug, PartialEq, Eq, Serialize, Deserialize)]
 pub struct PyManifestFileInfo {
+    #[pyo3(get)]
     pub id: String,
+    #[pyo3(get)]
     pub size_bytes: u64,
+    #[pyo3(get)]
     pub num_chunk_refs: u32,
 }
 
@@ -287,6 +291,7 @@ impl From<Diff> for PyDiff {
 
 #[pymethods]
 impl PyDiff {
+    #[allow(clippy::unwrap_used)]
     pub fn __repr__(&self) -> String {
         let mut res = String::new();
         use std::fmt::Write;
@@ -443,6 +448,29 @@ impl PyRepository {
         })
     }
 
+    #[classmethod]
+    #[pyo3(signature = (storage, *, config = None, authorize_virtual_chunk_access = None))]
+    fn create_async<'py>(
+        _cls: &Bound<'py, PyType>,
+        py: Python<'py>,
+        storage: PyStorage,
+        config: Option<&PyRepositoryConfig>,
+        authorize_virtual_chunk_access: Option<HashMap<String, Option<PyCredentials>>>,
+    ) -> PyResult<Bound<'py, PyAny>> {
+        let config =
+            config.map(|c| c.try_into().map_err(PyValueError::new_err)).transpose()?;
+        let authorize_virtual_chunk_access =
+            map_credentials(authorize_virtual_chunk_access);
+        pyo3_async_runtimes::tokio::future_into_py(py, async move {
+            let repository =
+                Repository::create(config, storage.0, authorize_virtual_chunk_access)
+                    .await
+                    .map_err(PyIcechunkStoreError::RepositoryError)?;
+
+            Ok(Self(Arc::new(RwLock::new(repository))))
+        })
+    }
+
     #[classmethod]
     #[pyo3(signature = (storage, *, config = None, authorize_virtual_chunk_access = None))]
     fn open(
@@ -472,6 +500,28 @@ impl PyRepository {
         })
     }
 
+    #[classmethod]
+    #[pyo3(signature = (storage, *, config = None, authorize_virtual_chunk_access = None))]
+    fn open_async<'py>(
+        _cls: &Bound<'py, PyType>,
+        py: Python<'py>,
+        storage: PyStorage,
+        config: Option<&PyRepositoryConfig>,
+        authorize_virtual_chunk_access: Option<HashMap<String, Option<PyCredentials>>>,
+    ) -> PyResult<Bound<'py, PyAny>> {
+        let config =
+            config.map(|c| c.try_into().map_err(PyValueError::new_err)).transpose()?;
+        let authorize_virtual_chunk_access =
+            map_credentials(authorize_virtual_chunk_access);
+        pyo3_async_runtimes::tokio::future_into_py(py, async move {
+            let repository =
+                Repository::open(config, storage.0, authorize_virtual_chunk_access)
+                    .await
+                    .map_err(PyIcechunkStoreError::RepositoryError)?;
+            Ok(Self(Arc::new(RwLock::new(repository))))
+        })
+    }
+
     #[classmethod]
     #[pyo3(signature = (storage, *, config = None, authorize_virtual_chunk_access = None))]
     fn open_or_create(
@@ -503,6 +553,31 @@ impl PyRepository {
         })
     }
 
+    #[classmethod]
+    #[pyo3(signature = (storage, *, config = None, authorize_virtual_chunk_access = None))]
+    fn open_or_create_async<'py>(
+        _cls: &Bound<'py, PyType>,
+        py: Python<'py>,
+        storage: PyStorage,
+        config: Option<&PyRepositoryConfig>,
+        authorize_virtual_chunk_access: Option<HashMap<String, Option<PyCredentials>>>,
+    ) -> PyResult<Bound<'py, PyAny>> {
+        let config =
+            config.map(|c| c.try_into().map_err(PyValueError::new_err)).transpose()?;
+        let authorize_virtual_chunk_access =
+            map_credentials(authorize_virtual_chunk_access);
+        pyo3_async_runtimes::tokio::future_into_py(py, async move {
+            let repository = Repository::open_or_create(
+                config,
+                storage.0,
+                authorize_virtual_chunk_access,
+            )
+            .await
+            .map_err(PyIcechunkStoreError::RepositoryError)?;
+            Ok(Self(Arc::new(RwLock::new(repository))))
+        })
+    }
+
     #[staticmethod]
     fn exists(py: Python<'_>, storage: PyStorage) -> PyResult<bool> {
         // This function calls block_on, so we need to allow other thread python to make progress
@@ -516,6 +591,16 @@ impl PyRepository {
         })
     }
 
+    #[staticmethod]
+    fn exists_async(py: Python<'_>, storage: PyStorage) -> PyResult<Bound<'_, PyAny>> {
+        pyo3_async_runtimes::tokio::future_into_py(py, async move {
+            let exists = Repository::exists(storage.0.as_ref())
+                .await
+                .map_err(PyIcechunkStoreError::RepositoryError)?;
+            Ok(exists)
+        })
+    }
+
     /// Reopen the repository changing its config and or virtual chunk credentials
     ///
     /// If config is None, it will use the same value as self
@@ -543,6 +628,30 @@ impl PyRepository {
         })
     }
 
+    #[pyo3(signature = (*, config = None, authorize_virtual_chunk_access = None::<Option<HashMap<String, Option<PyCredentials>>>>))]
+    fn reopen_async<'py>(
+        &'py self,
+        py: Python<'py>,
+        config: Option<&PyRepositoryConfig>,
+        authorize_virtual_chunk_access: Option<
+            Option<HashMap<String, Option<PyCredentials>>>,
+        >,
+    ) -> PyResult<Bound<'py, PyAny>> {
+        let existing_repository = self.0.clone();
+        let config =
+            config.map(|c| c.try_into().map_err(PyValueError::new_err)).transpose()?;
+        let authorize_virtual_chunk_access =
+            authorize_virtual_chunk_access.map(map_credentials);
+        pyo3_async_runtimes::tokio::future_into_py(py, async move {
+            let repository = existing_repository
+                .read()
+                .await
+                .reopen(config, authorize_virtual_chunk_access)
+                .map_err(PyIcechunkStoreError::RepositoryError)?;
+            Ok(Self(Arc::new(RwLock::new(repository))))
+        })
+    }
+
     #[classmethod]
     fn from_bytes(
         _cls: Bound<'_, PyType>,
@@ -557,7 +666,7 @@ impl PyRepository {
         })
     }
 
-    fn as_bytes(&self, py: Python<'_>) -> PyResult<Cow<[u8]>> {
+    fn as_bytes(&self, py: Python<'_>) -> PyResult<Cow<'_, [u8]>> {
         // This is a compute intensive task, we need to release the Gil
         py.allow_threads(move || {
             let bytes = self
@@ -585,6 +694,22 @@ impl PyRepository {
         })
     }
 
+    #[staticmethod]
+    fn fetch_config_async(
+        py: Python<'_>,
+        storage: PyStorage,
+    ) -> PyResult<Bound<'_, PyAny>> {
+        pyo3_async_runtimes::tokio::future_into_py::<_, Option<PyRepositoryConfig>>(
+            py,
+            async move {
+                let res = Repository::fetch_config(storage.0.as_ref())
+                    .await
+                    .map_err(PyIcechunkStoreError::RepositoryError)?;
+                Ok(res.map(|res| res.0.into()))
+            },
+        )
+    }
+
     fn save_config(&self, py: Python<'_>) -> PyResult<()> {
         // This function calls block_on, so we need to allow other thread python to make progress
         py.allow_threads(move || {
@@ -601,6 +726,18 @@ impl PyRepository {
         })
     }
 
+    fn save_config_async<'py>(&'py self, py: Python<'py>) -> PyResult<Bound<'py, PyAny>> {
+        let repository = self.0.clone();
+        pyo3_async_runtimes::tokio::future_into_py(py, async move {
+            let repository = repository.read().await;
+            repository
+                .save_config()
+                .await
+                .map_err(PyIcechunkStoreError::RepositoryError)?;
+            Ok(())
+        })
+    }
+
     pub fn config(&self) -> PyRepositoryConfig {
         self.0.blocking_read().config().clone().into()
     }
@@ -613,6 +750,15 @@ impl PyRepository {
         PyStorage(Arc::clone(self.0.blocking_read().storage()))
     }
 
+    #[getter]
+    fn authorized_virtual_container_prefixes<'py>(
+        &self,
+        py: Python<'py>,
+    ) -> PyResult<Bound<'py, PySet>> {
+        let prefixes = self.0.blocking_read().authorized_virtual_container_prefixes();
+        PySet::new(py, prefixes.iter().map(|s| s.as_str()))
+    }
+
     pub fn set_default_commit_metadata(
         &self,
         py: Python<'_>,
@@ -695,6 +841,30 @@ impl PyRepository {
         })
     }
 
+    fn create_branch_async<'py>(
+        &'py self,
+        py: Python<'py>,
+        branch_name: &str,
+        snapshot_id: &str,
+    ) -> PyResult<Bound<'py, PyAny>> {
+        let repository = self.0.clone();
+        let branch_name = branch_name.to_owned();
+        let snapshot_id = SnapshotId::try_from(snapshot_id).map_err(|_| {
+            PyIcechunkStoreError::RepositoryError(
+                RepositoryErrorKind::InvalidSnapshotId(snapshot_id.to_owned()).into(),
+            )
+        })?;
+
+        pyo3_async_runtimes::tokio::future_into_py(py, async move {
+            let repository = repository.read().await;
+            repository
+                .create_branch(&branch_name, &snapshot_id)
+                .await
+                .map_err(PyIcechunkStoreError::RepositoryError)?;
+            Ok(())
+        })
+    }
+
     pub fn list_branches(&self, py: Python<'_>) -> PyResult<BTreeSet<String>> {
         // This function calls block_on, so we need to allow other thread python to make progress
         py.allow_threads(move || {
@@ -711,6 +881,24 @@ impl PyRepository {
         })
     }
 
+    fn list_branches_async<'py>(
+        &'py self,
+        py: Python<'py>,
+    ) -> PyResult<Bound<'py, PyAny>> {
+        let repository = self.0.clone();
+        pyo3_async_runtimes::tokio::future_into_py::<_, BTreeSet<String>>(
+            py,
+            async move {
+                let repository = repository.read().await;
+                let branches = repository
+                    .list_branches()
+                    .await
+                    .map_err(PyIcechunkStoreError::RepositoryError)?;
+                Ok(branches)
+            },
+        )
+    }
+
     pub fn lookup_branch(&self, py: Python<'_>, branch_name: &str) -> PyResult<String> {
         // This function calls block_on, so we need to allow other thread python to make progress
         py.allow_threads(move || {
@@ -727,6 +915,23 @@ impl PyRepository {
         })
     }
 
+    fn lookup_branch_async<'py>(
+        &'py self,
+        py: Python<'py>,
+        branch_name: &str,
+    ) -> PyResult<Bound<'py, PyAny>> {
+        let repository = self.0.clone();
+        let branch_name = branch_name.to_owned();
+        pyo3_async_runtimes::tokio::future_into_py::<_, String>(py, async move {
+            let repository = repository.read().await;
+            let tip = repository
+                .lookup_branch(&branch_name)
+                .await
+                .map_err(PyIcechunkStoreError::RepositoryError)?;
+            Ok(tip.to_string())
+        })
+    }
+
     pub fn lookup_snapshot(
         &self,
         py: Python<'_>,
@@ -752,25 +957,58 @@ impl PyRepository {
         })
     }
 
+    fn lookup_snapshot_async<'py>(
+        &'py self,
+        py: Python<'py>,
+        snapshot_id: &str,
+    ) -> PyResult<Bound<'py, PyAny>> {
+        let repository = self.0.clone();
+        let snapshot_id = SnapshotId::try_from(snapshot_id).map_err(|_| {
+            PyIcechunkStoreError::RepositoryError(
+                RepositoryErrorKind::InvalidSnapshotId(snapshot_id.to_owned()).into(),
+            )
+        })?;
+        pyo3_async_runtimes::tokio::future_into_py::<_, PySnapshotInfo>(py, async move {
+            let repository = repository.read().await;
+            let res = repository
+                .lookup_snapshot(&snapshot_id)
+                .await
+                .map_err(PyIcechunkStoreError::RepositoryError)?;
+            Ok(res.into())
+        })
+    }
+
     pub fn reset_branch(
         &self,
         py: Python<'_>,
         branch_name: &str,
-        snapshot_id: &str,
+        to_snapshot_id: &str,
+        from_snapshot_id: Option<&str>,
     ) -> PyResult<()> {
         // This function calls block_on, so we need to allow other thread python to make progress
         py.allow_threads(move || {
-            let snapshot_id = SnapshotId::try_from(snapshot_id).map_err(|_| {
+            let to_snapshot_id = SnapshotId::try_from(to_snapshot_id).map_err(|_| {
                 PyIcechunkStoreError::RepositoryError(
-                    RepositoryErrorKind::InvalidSnapshotId(snapshot_id.to_owned()).into(),
+                    RepositoryErrorKind::InvalidSnapshotId(to_snapshot_id.to_owned())
+                        .into(),
                 )
             })?;
 
+            let from_snapshot_id = from_snapshot_id
+                .map(|sid| {
+                    SnapshotId::try_from(sid).map_err(|_| {
+                        PyIcechunkStoreError::RepositoryError(
+                            RepositoryErrorKind::InvalidSnapshotId(sid.to_owned()).into(),
+                        )
+                    })
+                })
+                .transpose()?;
+
             pyo3_async_runtimes::tokio::get_runtime().block_on(async move {
                 self.0
                     .read()
                     .await
-                    .reset_branch(branch_name, &snapshot_id)
+                    .reset_branch(branch_name, &to_snapshot_id, from_snapshot_id.as_ref())
                     .await
                     .map_err(PyIcechunkStoreError::RepositoryError)?;
                 Ok(())
@@ -778,6 +1016,41 @@ impl PyRepository {
         })
     }
 
+    fn reset_branch_async<'py>(
+        &'py self,
+        py: Python<'py>,
+        branch_name: &str,
+        to_snapshot_id: &str,
+        from_snapshot_id: Option<&str>,
+    ) -> PyResult<Bound<'py, PyAny>> {
+        let repository = self.0.clone();
+        let branch_name = branch_name.to_owned();
+        let to_snapshot_id = SnapshotId::try_from(to_snapshot_id).map_err(|_| {
+            PyIcechunkStoreError::RepositoryError(
+                RepositoryErrorKind::InvalidSnapshotId(to_snapshot_id.to_owned()).into(),
+            )
+        })?;
+
+        let from_snapshot_id = from_snapshot_id
+            .map(|sid| {
+                SnapshotId::try_from(sid).map_err(|_| {
+                    PyIcechunkStoreError::RepositoryError(
+                        RepositoryErrorKind::InvalidSnapshotId(sid.to_owned()).into(),
+                    )
+                })
+            })
+            .transpose()?;
+
+        pyo3_async_runtimes::tokio::future_into_py(py, async move {
+            let repository = repository.read().await;
+            repository
+                .reset_branch(&branch_name, &to_snapshot_id, from_snapshot_id.as_ref())
+                .await
+                .map_err(PyIcechunkStoreError::RepositoryError)?;
+            Ok(())
+        })
+    }
+
     pub fn delete_branch(&self, py: Python<'_>, branch: &str) -> PyResult<()> {
         // This function calls block_on, so we need to allow other thread python to make progress
         py.allow_threads(move || {
@@ -793,6 +1066,23 @@ impl PyRepository {
         })
     }
 
+    fn delete_branch_async<'py>(
+        &'py self,
+        py: Python<'py>,
+        branch: &str,
+    ) -> PyResult<Bound<'py, PyAny>> {
+        let repository = self.0.clone();
+        let branch = branch.to_owned();
+        pyo3_async_runtimes::tokio::future_into_py(py, async move {
+            let repository = repository.read().await;
+            repository
+                .delete_branch(&branch)
+                .await
+                .map_err(PyIcechunkStoreError::RepositoryError)?;
+            Ok(())
+        })
+    }
+
     pub fn delete_tag(&self, py: Python<'_>, tag: &str) -> PyResult<()> {
         // This function calls block_on, so we need to allow other thread python to make progress
         py.allow_threads(move || {
@@ -808,6 +1098,23 @@ impl PyRepository {
         })
     }
 
+    fn delete_tag_async<'py>(
+        &'py self,
+        py: Python<'py>,
+        tag: &str,
+    ) -> PyResult<Bound<'py, PyAny>> {
+        let repository = self.0.clone();
+        let tag = tag.to_owned();
+        pyo3_async_runtimes::tokio::future_into_py(py, async move {
+            let repository = repository.read().await;
+            repository
+                .delete_tag(&tag)
+                .await
+                .map_err(PyIcechunkStoreError::RepositoryError)?;
+            Ok(())
+        })
+    }
+
     pub fn create_tag(
         &self,
         py: Python<'_>,
@@ -834,6 +1141,30 @@ impl PyRepository {
         })
     }
 
+    fn create_tag_async<'py>(
+        &'py self,
+        py: Python<'py>,
+        tag_name: &str,
+        snapshot_id: &str,
+    ) -> PyResult<Bound<'py, PyAny>> {
+        let repository = self.0.clone();
+        let tag_name = tag_name.to_owned();
+        let snapshot_id = SnapshotId::try_from(snapshot_id).map_err(|_| {
+            PyIcechunkStoreError::RepositoryError(
+                RepositoryErrorKind::InvalidSnapshotId(snapshot_id.to_owned()).into(),
+            )
+        })?;
+
+        pyo3_async_runtimes::tokio::future_into_py(py, async move {
+            let repository = repository.read().await;
+            repository
+                .create_tag(&tag_name, &snapshot_id)
+                .await
+                .map_err(PyIcechunkStoreError::RepositoryError)?;
+            Ok(())
+        })
+    }
+
     pub fn list_tags(&self, py: Python<'_>) -> PyResult<BTreeSet<String>> {
         // This function calls block_on, so we need to allow other thread python to make progress
         py.allow_threads(move || {
@@ -850,6 +1181,21 @@ impl PyRepository {
         })
     }
 
+    fn list_tags_async<'py>(&'py self, py: Python<'py>) -> PyResult<Bound<'py, PyAny>> {
+        let repository = self.0.clone();
+        pyo3_async_runtimes::tokio::future_into_py::<_, BTreeSet<String>>(
+            py,
+            async move {
+                let repository = repository.read().await;
+                let tags = repository
+                    .list_tags()
+                    .await
+                    .map_err(PyIcechunkStoreError::RepositoryError)?;
+                Ok(tags)
+            },
+        )
+    }
+
     pub fn lookup_tag(&self, py: Python<'_>, tag: &str) -> PyResult<String> {
         // This function calls block_on, so we need to allow other thread python to make progress
         py.allow_threads(move || {
@@ -866,6 +1212,23 @@ impl PyRepository {
         })
     }
 
+    fn lookup_tag_async<'py>(
+        &'py self,
+        py: Python<'py>,
+        tag: &str,
+    ) -> PyResult<Bound<'py, PyAny>> {
+        let repository = self.0.clone();
+        let tag_name = tag.to_owned();
+        pyo3_async_runtimes::tokio::future_into_py::<_, String>(py, async move {
+            let repository = repository.read().await;
+            let tip = repository
+                .lookup_tag(&tag_name)
+                .await
+                .map_err(PyIcechunkStoreError::RepositoryError)?;
+            Ok(tip.to_string())
+        })
+    }
+
     #[pyo3(signature = (*, from_branch=None, from_tag=None, from_snapshot_id=None, to_branch=None, to_tag=None, to_snapshot_id=None))]
     #[allow(clippy::too_many_arguments)]
     pub fn diff(
@@ -896,6 +1259,32 @@ impl PyRepository {
         })
     }
 
+    #[pyo3(signature = (*, from_branch=None, from_tag=None, from_snapshot_id=None, to_branch=None, to_tag=None, to_snapshot_id=None))]
+    #[allow(clippy::too_many_arguments)]
+    fn diff_async<'py>(
+        &'py self,
+        py: Python<'py>,
+        from_branch: Option<String>,
+        from_tag: Option<String>,
+        from_snapshot_id: Option<String>,
+        to_branch: Option<String>,
+        to_tag: Option<String>,
+        to_snapshot_id: Option<String>,
+    ) -> PyResult<Bound<'py, PyAny>> {
+        let from = args_to_version_info(from_branch, from_tag, from_snapshot_id, None)?;
+        let to = args_to_version_info(to_branch, to_tag, to_snapshot_id, None)?;
+        let repository = self.0.clone();
+
+        pyo3_async_runtimes::tokio::future_into_py::<_, PyDiff>(py, async move {
+            let repository = repository.read().await;
+            let diff = repository
+                .diff(&from, &to)
+                .await
+                .map_err(PyIcechunkStoreError::SessionError)?;
+            Ok(diff.into())
+        })
+    }
+
     #[pyo3(signature = (*, branch = None, tag = None, snapshot_id = None, as_of = None))]
     pub fn readonly_session(
         &self,
@@ -922,6 +1311,28 @@ impl PyRepository {
         })
     }
 
+    #[pyo3(signature = (*, branch = None, tag = None, snapshot_id = None, as_of = None))]
+    fn readonly_session_async<'py>(
+        &'py self,
+        py: Python<'py>,
+        branch: Option<String>,
+        tag: Option<String>,
+        snapshot_id: Option<String>,
+        as_of: Option<DateTime<Utc>>,
+    ) -> PyResult<Bound<'py, PyAny>> {
+        let version = args_to_version_info(branch, tag, snapshot_id, as_of)?;
+        let repository = self.0.clone();
+
+        pyo3_async_runtimes::tokio::future_into_py::<_, PySession>(py, async move {
+            let repository = repository.read().await;
+            let session = repository
+                .readonly_session(&version)
+                .await
+                .map_err(PyIcechunkStoreError::RepositoryError)?;
+            Ok(PySession(Arc::new(RwLock::new(session))))
+        })
+    }
+
     pub fn writable_session(&self, py: Python<'_>, branch: &str) -> PyResult<PySession> {
         // This function calls block_on, so we need to allow other thread python to make progress
         py.allow_threads(move || {
@@ -939,6 +1350,23 @@ impl PyRepository {
         })
     }
 
+    fn writable_session_async<'py>(
+        &'py self,
+        py: Python<'py>,
+        branch: &str,
+    ) -> PyResult<Bound<'py, PyAny>> {
+        let repository = self.0.clone();
+        let branch = branch.to_owned();
+        pyo3_async_runtimes::tokio::future_into_py::<_, PySession>(py, async move {
+            let repository = repository.read().await;
+            let session = repository
+                .writable_session(&branch)
+                .await
+                .map_err(PyIcechunkStoreError::RepositoryError)?;
+            Ok(PySession(Arc::new(RwLock::new(session))))
+        })
+    }
+
     #[pyo3(signature = (message, branch, metadata=None))]
     pub fn rewrite_manifests(
         &self,
@@ -961,6 +1389,27 @@ impl PyRepository {
         })
     }
 
+    #[pyo3(signature = (message, branch, metadata=None))]
+    fn rewrite_manifests_async<'py>(
+        &'py self,
+        py: Python<'py>,
+        message: &str,
+        branch: &str,
+        metadata: Option<PySnapshotProperties>,
+    ) -> PyResult<Bound<'py, PyAny>> {
+        let repository = self.0.clone();
+        let message = message.to_owned();
+        let branch = branch.to_owned();
+        let metadata = metadata.map(|m| m.into());
+        pyo3_async_runtimes::tokio::future_into_py::<_, String>(py, async move {
+            let repository = repository.read().await;
+            let result = rewrite_manifests(&repository, &branch, &message, metadata)
+                .await
+                .map_err(PyIcechunkStoreError::ManifestOpsError)?;
+            Ok(result.to_string())
+        })
+    }
+
     #[pyo3(signature = (older_than, *, delete_expired_branches = false, delete_expired_tags = false))]
     pub fn expire_snapshots(
         &self,
@@ -1013,6 +1462,47 @@ impl PyRepository {
         })
     }
 
+    #[pyo3(signature = (older_than, *, delete_expired_branches = false, delete_expired_tags = false))]
+    fn expire_snapshots_async<'py>(
+        &'py self,
+        py: Python<'py>,
+        older_than: DateTime<Utc>,
+        delete_expired_branches: bool,
+        delete_expired_tags: bool,
+    ) -> PyResult<Bound<'py, PyAny>> {
+        let repository = self.0.clone();
+        pyo3_async_runtimes::tokio::future_into_py::<_, HashSet<String>>(py, async move {
+            let (storage, storage_settings, asset_manager) = {
+                let lock = repository.read().await;
+                (
+                    Arc::clone(lock.storage()),
+                    lock.storage_settings().clone(),
+                    Arc::clone(lock.asset_manager()),
+                )
+            };
+
+            let result = expire(
+                storage.as_ref(),
+                &storage_settings,
+                asset_manager,
+                older_than,
+                if delete_expired_branches {
+                    ExpiredRefAction::Delete
+                } else {
+                    ExpiredRefAction::Ignore
+                },
+                if delete_expired_tags {
+                    ExpiredRefAction::Delete
+                } else {
+                    ExpiredRefAction::Ignore
+                },
+            )
+            .await
+            .map_err(PyIcechunkStoreError::GCError)?;
+            Ok(result.released_snapshots.iter().map(|id| id.to_string()).collect())
+        })
+    }
+
     pub fn garbage_collect(
         &self,
         py: Python<'_>,
@@ -1058,6 +1548,46 @@ impl PyRepository {
         })
     }
 
+    fn garbage_collect_async<'py>(
+        &'py self,
+        py: Python<'py>,
+        delete_object_older_than: DateTime<Utc>,
+        dry_run: bool,
+        max_snapshots_in_memory: NonZeroU16,
+        max_compressed_manifest_mem_bytes: NonZeroUsize,
+        max_concurrent_manifest_fetches: NonZeroU16,
+    ) -> PyResult<Bound<'py, PyAny>> {
+        let repository = self.0.clone();
+        pyo3_async_runtimes::tokio::future_into_py::<_, PyGCSummary>(py, async move {
+            let gc_config = GCConfig::clean_all(
+                delete_object_older_than,
+                delete_object_older_than,
+                Default::default(),
+                max_snapshots_in_memory,
+                max_compressed_manifest_mem_bytes,
+                max_concurrent_manifest_fetches,
+                dry_run,
+            );
+            let (storage, storage_settings, asset_manager) = {
+                let lock = repository.read().await;
+                (
+                    Arc::clone(lock.storage()),
+                    lock.storage_settings().clone(),
+                    Arc::clone(lock.asset_manager()),
+                )
+            };
+            let result = garbage_collect(
+                storage.as_ref(),
+                &storage_settings,
+                asset_manager,
+                &gc_config,
+            )
+            .await
+            .map_err(PyIcechunkStoreError::GCError)?;
+            Ok(result.into())
+        })
+    }
+
     pub fn total_chunks_storage(
         &self,
         py: Python<'_>,
@@ -1093,6 +1623,73 @@ impl PyRepository {
             Ok(result)
         })
     }
+
+    fn total_chunks_storage_async<'py>(
+        &'py self,
+        py: Python<'py>,
+        max_snapshots_in_memory: NonZeroU16,
+        max_compressed_manifest_mem_bytes: NonZeroUsize,
+        max_concurrent_manifest_fetches: NonZeroU16,
+    ) -> PyResult<Bound<'py, PyAny>> {
+        let repository = self.0.clone();
+        pyo3_async_runtimes::tokio::future_into_py::<_, u64>(py, async move {
+            let (storage, storage_settings, asset_manager) = {
+                let lock = repository.read().await;
+                (
+                    Arc::clone(lock.storage()),
+                    lock.storage_settings().clone(),
+                    Arc::clone(lock.asset_manager()),
+                )
+            };
+            let result = repo_chunks_storage(
+                storage.as_ref(),
+                &storage_settings,
+                asset_manager,
+                max_snapshots_in_memory,
+                max_compressed_manifest_mem_bytes,
+                max_concurrent_manifest_fetches,
+            )
+            .await
+            .map_err(PyIcechunkStoreError::RepositoryError)?;
+            Ok(result)
+        })
+    }
+
+    #[pyo3(signature = (snapshot_id, *, pretty = true))]
+    fn inspect_snapshot(&self, snapshot_id: String, pretty: bool) -> PyResult<String> {
+        let result = pyo3_async_runtimes::tokio::get_runtime()
+            .block_on(async move {
+                let lock = self.0.read().await;
+                let snap = SnapshotId::try_from(snapshot_id.as_str())
+                    .map_err(|e| RepositoryErrorKind::Other(e.to_string()))?;
+                let res = snapshot_json(lock.asset_manager(), &snap, pretty).await?;
+                Ok(res)
+            })
+            .map_err(PyIcechunkStoreError::RepositoryError)?;
+        Ok(result)
+    }
+
+    #[pyo3(signature = (snapshot_id, *, pretty = true))]
+    fn inspect_snapshot_async<'py>(
+        &self,
+        py: Python<'py>,
+        snapshot_id: String,
+        pretty: bool,
+    ) -> PyResult<Bound<'py, PyAny>> {
+        let repository = self.0.clone();
+        pyo3_async_runtimes::tokio::future_into_py(py, async move {
+            let lock = repository.read().await;
+            let snap = SnapshotId::try_from(snapshot_id.as_str())
+                .map_err(|e| {
+                    RepositoryError::from(RepositoryErrorKind::Other(e.to_string()))
+                })
+                .map_err(PyIcechunkStoreError::RepositoryError)?;
+            let res = snapshot_json(lock.asset_manager(), &snap, pretty)
+                .await
+                .map_err(PyIcechunkStoreError::RepositoryError)?;
+            Ok(res)
+        })
+    }
 }
 
 fn map_credentials(
@@ -1140,8 +1737,6 @@ fn args_to_version_info(
 
         Ok(VersionInfo::SnapshotId(snapshot_id))
     } else {
-        return Err(PyValueError::new_err(
-            "Must provide one of branch, tag, or snapshot_id",
-        ));
+        Err(PyValueError::new_err("Must provide one of branch, tag, or snapshot_id"))
     }
 }
