@@ -1,6 +1,5 @@
 use std::{borrow::Cow, sync::Arc};
 
-use bytes::Bytes;
 use chrono::Utc;
 use futures::{StreamExt, TryStreamExt};
 use icechunk::{
@@ -18,6 +17,7 @@ use pyo3::{
     prelude::*,
     types::{PyTuple, PyType},
 };
+use pyo3_bytes::PyBytes;
 use serde::{Deserialize, Serialize};
 use tokio::sync::Mutex;
 
@@ -105,11 +105,11 @@ impl PyStore {
     fn from_bytes(
         _cls: Bound<'_, PyType>,
         py: Python<'_>,
-        bytes: Vec<u8>,
+        bytes: PyBytes,
     ) -> PyResult<Self> {
         // This is a compute intensive task, we need to release the Gil
         py.allow_threads(move || {
-            let bytes = Bytes::from(bytes);
+            let bytes = bytes.into_inner();
             let store = Store::from_bytes(bytes).map_err(|e| {
                 PyValueError::new_err(format!(
                     "Failed to deserialize store from bytes: {e}"
@@ -135,7 +135,7 @@ impl PyStore {
         })
     }
 
-    fn as_bytes(&self, py: Python<'_>) -> PyResult<Cow<[u8]>> {
+    fn as_bytes(&self, py: Python<'_>) -> PyResult<Cow<'_, [u8]>> {
         // This is blocking function, we need to release the Gil
         py.allow_threads(move || {
             // FIXME: Use rmp_serde instead of serde_json to optimize performance
@@ -199,7 +199,7 @@ impl PyStore {
             // We need to distinguish the "safe" case of trying to fetch an uninitialized key
             // from other types of errors, we use PyKeyError exception for that
             match data {
-                Ok(data) => Ok(Vec::from(data)),
+                Ok(data) => Ok(PyBytes::new(data)),
                 Err(StoreError { kind: StoreErrorKind::NotFound(_), .. }) => {
                     Err(PyKeyError::new_err(key))
                 }
@@ -267,7 +267,7 @@ impl PyStore {
         &'py self,
         py: Python<'py>,
         key: String,
-        value: Vec<u8>,
+        value: PyBytes,
     ) -> PyResult<Bound<'py, PyAny>> {
         let store = Arc::clone(&self.0);
 
@@ -279,7 +279,7 @@ impl PyStore {
         // see this tracking issue: https://github.com/PyO3/pyo3/issues/1632
         pyo3_async_runtimes::tokio::future_into_py(py, async move {
             store
-                .set(&key, Bytes::from(value))
+                .set(&key, value.into_inner())
                 .await
                 .map_err(PyIcechunkStoreError::from)?;
             Ok(())
@@ -290,13 +290,13 @@ impl PyStore {
         &'py self,
         py: Python<'py>,
         key: String,
-        value: Vec<u8>,
+        value: PyBytes,
     ) -> PyResult<Bound<'py, PyAny>> {
         let store = Arc::clone(&self.0);
 
         pyo3_async_runtimes::tokio::future_into_py(py, async move {
             store
-                .set_if_not_exists(&key, Bytes::from(value))
+                .set_if_not_exists(&key, value.into_inner())
                 .await
                 .map_err(PyIcechunkStoreError::from)?;
             Ok(())
@@ -304,7 +304,6 @@ impl PyStore {
     }
 
     #[allow(clippy::too_many_arguments)]
-    #[pyo3(signature = (key, location, offset, length, checksum = None, validate_container = false))]
     fn set_virtual_ref(
         &self,
         py: Python<'_>,
@@ -320,8 +319,11 @@ impl PyStore {
             let store = Arc::clone(&self.0);
 
             pyo3_async_runtimes::tokio::get_runtime().block_on(async move {
+                let location =
+                    VirtualChunkLocation::from_absolute_path(location.as_str())
+                        .map_err(PyIcechunkStoreError::from)?;
                 let virtual_ref = VirtualChunkRef {
-                    location: VirtualChunkLocation(location),
+                    location,
                     offset,
                     length,
                     checksum: checksum.map(|cs| cs.into()),
@@ -335,6 +337,35 @@ impl PyStore {
         })
     }
 
+    #[allow(clippy::too_many_arguments)]
+    fn set_virtual_ref_async<'py>(
+        &'py self,
+        py: Python<'py>,
+        key: String,
+        location: String,
+        offset: ChunkOffset,
+        length: ChunkLength,
+        checksum: Option<ChecksumArgument>,
+        validate_container: bool,
+    ) -> PyResult<Bound<'py, PyAny>> {
+        let store = Arc::clone(&self.0);
+        pyo3_async_runtimes::tokio::future_into_py(py, async move {
+            let location = VirtualChunkLocation::from_absolute_path(location.as_str())
+                .map_err(PyIcechunkStoreError::from)?;
+            let virtual_ref = VirtualChunkRef {
+                location,
+                offset,
+                length,
+                checksum: checksum.map(|cs| cs.into()),
+            };
+            store
+                .set_virtual_ref(&key, virtual_ref, validate_container)
+                .await
+                .map_err(PyIcechunkStoreError::from)?;
+            Ok(())
+        })
+    }
+
     fn set_virtual_refs(
         &self,
         py: Python<'_>,
@@ -346,17 +377,24 @@ impl PyStore {
             let store = Arc::clone(&self.0);
 
             let res = pyo3_async_runtimes::tokio::get_runtime().block_on(async move {
-                let vrefs = chunks.into_iter().map(|vcs| {
-                    let checksum = vcs.checksum();
-                    let index = ChunkIndices(vcs.index);
-                    let vref = VirtualChunkRef {
-                        location: VirtualChunkLocation(vcs.location),
-                        offset: vcs.offset,
-                        length: vcs.length,
-                        checksum,
-                    };
-                    (index, vref)
-                });
+                let vrefs: Vec<(ChunkIndices, VirtualChunkRef)> = chunks
+                    .into_iter()
+                    .map(|vcs| {
+                        let checksum = vcs.checksum();
+                        let index = ChunkIndices(vcs.index);
+                        let location = VirtualChunkLocation::from_absolute_path(
+                            vcs.location.as_str(),
+                        )
+                        .map_err(PyIcechunkStoreError::from)?;
+                        let vref = VirtualChunkRef {
+                            offset: vcs.offset,
+                            length: vcs.length,
+                            location,
+                            checksum,
+                        };
+                        Ok::<_, PyIcechunkStoreError>((index, vref))
+                    })
+                    .try_collect()?;
 
                 let array_path = if !array_path.starts_with("/") {
                     format!("/{array_path}")
@@ -389,6 +427,66 @@ impl PyStore {
         })
     }
 
+    fn set_virtual_refs_async<'py>(
+        &'py self,
+        py: Python<'py>,
+        array_path: String,
+        chunks: Vec<VirtualChunkSpec>,
+        validate_containers: bool,
+    ) -> PyResult<Bound<'py, PyAny>> {
+        let store = Arc::clone(&self.0);
+        pyo3_async_runtimes::tokio::future_into_py::<_, Option<Vec<Py<PyTuple>>>>(
+            py,
+            async move {
+                let vrefs: Vec<(ChunkIndices, VirtualChunkRef)> = chunks
+                    .into_iter()
+                    .map(|vcs| {
+                        let checksum = vcs.checksum();
+                        let index = ChunkIndices(vcs.index);
+                        let location = VirtualChunkLocation::from_absolute_path(
+                            vcs.location.as_str(),
+                        )
+                        .map_err(PyIcechunkStoreError::from)?;
+                        let vref = VirtualChunkRef {
+                            offset: vcs.offset,
+                            length: vcs.length,
+                            location,
+                            checksum,
+                        };
+                        Ok::<_, PyIcechunkStoreError>((index, vref))
+                    })
+                    .try_collect()?;
+
+                let array_path = if !array_path.starts_with("/") {
+                    format!("/{array_path}")
+                } else {
+                    array_path.to_string()
+                };
+
+                let path = Path::try_from(array_path).map_err(|e| {
+                    PyValueError::new_err(format!("Invalid array path: {e}"))
+                })?;
+
+                let res = store
+                    .set_virtual_refs(&path, validate_containers, vrefs)
+                    .await
+                    .map_err(PyIcechunkStoreError::from)?;
+
+                match res {
+                    SetVirtualRefsResult::Done => Ok(None),
+                    SetVirtualRefsResult::FailedRefs(vec) => Python::with_gil(|py| {
+                        let res = vec
+                            .into_iter()
+                            .map(|ci| PyTuple::new(py, ci.0).map(|tup| tup.unbind()))
+                            .try_collect()?;
+
+                        Ok(Some(res))
+                    }),
+                }
+            },
+        )
+    }
+
     fn delete<'py>(
         &'py self,
         py: Python<'py>,
@@ -424,7 +522,7 @@ impl PyStore {
     fn set_partial_values<'py>(
         &'py self,
         py: Python<'py>,
-        key_start_values: Vec<(String, ChunkOffset, Vec<u8>)>,
+        key_start_values: Vec<(String, ChunkOffset, PyBytes)>,
     ) -> PyResult<Bound<'py, PyAny>> {
         // We need to get our own copy of the keys to pass to the downstream store function because that
         // function requires a Vec<&str, which we cannot borrow from when we are borrowing from the tuple
@@ -445,7 +543,7 @@ impl PyStore {
         pyo3_async_runtimes::tokio::future_into_py(py, async move {
             let mapped_to_bytes = key_start_values.into_iter().enumerate().map(
                 |(i, (_key, offset, value))| {
-                    (keys[i].as_str(), offset, Bytes::from(value))
+                    (keys[i].as_str(), offset, value.into_inner())
                 },
             );
 
