@@ -320,6 +320,7 @@ fn add_repo_to_config(
                     anonymous: false,
                     allow_http: false,
                     force_path_style: false,
+                    network_stream_timeout_seconds: None,
                 },
                 credentials: S3Credentials::FromEnv,
                 config: RepositoryConfig::default(),
@@ -351,6 +352,7 @@ fn add_repo_to_config(
                     anonymous: false,
                     allow_http: false,
                     force_path_style: false,
+                    network_stream_timeout_seconds: None,
                 },
                 credentials: S3Credentials::FromEnv,
                 config: RepositoryConfig::default(),
@@ -432,7 +434,7 @@ pub async fn run_cli(args: IcechunkCLI) -> Result<()> {
         }
         Command::Config(ConfigCommand::List) => config_list(&config, stdout()).await,
     }
-    .map_err(|e| anyhow::anyhow!("❌ {}", e))
+    .map_err(|e| anyhow::anyhow!("❌ {e}"))
 }
 
 #[cfg(test)]
