@@ -24,6 +24,7 @@ pub struct S3Options {
     // field was added in v0.2.6
     #[serde(default = "default_force_path_style")]
     pub force_path_style: bool,
+    pub network_stream_timeout_seconds: Option<u32>,
 }
 
 fn default_force_path_style() -> bool {
@@ -34,12 +35,15 @@ impl fmt::Display for S3Options {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         write!(
             f,
-            "S3Options(region={}, endpoint_url={}, anonymous={}, allow_http={}, force_path_style={})",
+            "S3Options(region={}, endpoint_url={}, anonymous={}, allow_http={}, force_path_style={}, network_stream_timeout_seconds={})",
             self.region.as_deref().unwrap_or("None"),
             self.endpoint_url.as_deref().unwrap_or("None"),
             self.anonymous,
             self.allow_http,
             self.force_path_style,
+            self.network_stream_timeout_seconds
+                .map(|n| n.to_string())
+                .unwrap_or("None".to_string()),
         )
     }
 }
@@ -98,10 +102,10 @@ pub struct CachingConfig {
 
 impl CachingConfig {
     pub fn num_snapshot_nodes(&self) -> u64 {
-        self.num_snapshot_nodes.unwrap_or(10_000)
+        self.num_snapshot_nodes.unwrap_or(500_000)
     }
     pub fn num_chunk_refs(&self) -> u64 {
-        self.num_chunk_refs.unwrap_or(5_000_000)
+        self.num_chunk_refs.unwrap_or(15_000_000)
     }
     pub fn num_transaction_changes(&self) -> u64 {
         self.num_transaction_changes.unwrap_or(0)
@@ -354,6 +358,7 @@ pub struct RepositoryConfig {
     pub get_partial_values_concurrency: Option<u16>,
 
     pub compression: Option<CompressionConfig>,
+    pub max_concurrent_requests: Option<u16>,
     pub caching: Option<CachingConfig>,
 
     // If not set it will use the Storage implementation default
@@ -371,6 +376,7 @@ pub struct RepositoryConfig {
 static DEFAULT_COMPRESSION: OnceLock<CompressionConfig> = OnceLock::new();
 static DEFAULT_CACHING: OnceLock<CachingConfig> = OnceLock::new();
 static DEFAULT_MANIFEST_CONFIG: OnceLock<ManifestConfig> = OnceLock::new();
+pub const DEFAULT_MAX_CONCURRENT_REQUESTS: u16 = 256;
 
 impl RepositoryConfig {
     pub fn inline_chunk_threshold_bytes(&self) -> u16 {
@@ -400,6 +406,10 @@ impl RepositoryConfig {
         })
     }
 
+    pub fn max_concurrent_requests(&self) -> u16 {
+        self.max_concurrent_requests.unwrap_or(DEFAULT_MAX_CONCURRENT_REQUESTS)
+    }
+
     pub fn merge(&self, other: Self) -> Self {
         Self {
             inline_chunk_threshold_bytes: other
@@ -414,6 +424,15 @@ impl RepositoryConfig {
                 (Some(c), None) => Some(c.clone()),
                 (Some(mine), Some(theirs)) => Some(mine.merge(theirs)),
             },
+            max_concurrent_requests: match (
+                &self.max_concurrent_requests,
+                other.max_concurrent_requests,
+            ) {
+                (None, None) => None,
+                (None, Some(c)) => Some(c),
+                (Some(c), None) => Some(*c),
+                (Some(_), Some(theirs)) => Some(theirs),
+            },
             caching: match (&self.caching, other.caching) {
                 (None, None) => None,
                 (None, Some(c)) => Some(c),
@@ -503,7 +522,8 @@ pub enum S3Credentials {
 }
 
 #[derive(Clone, Debug, Deserialize, Serialize, PartialEq, Eq)]
-#[serde(tag = "gcs_static_credential_type")]
+// We need to adjacently tag because we more than one variant with matching inner types https://github.com/serde-rs/serde/issues/1307
+#[serde(tag = "gcs_static_credential_type", content = "__field0")]
 #[serde(rename_all = "snake_case")]
 pub enum GcsStaticCredentials {
     ServiceAccount(PathBuf),
@@ -542,7 +562,8 @@ pub enum GcsCredentials {
 }
 
 #[derive(Clone, Debug, Deserialize, Serialize, PartialEq, Eq)]
-#[serde(tag = "az_static_credential_type")]
+// We need to adjacently tag because we more than one variant with matching inner types https://github.com/serde-rs/serde/issues/1307
+#[serde(tag = "az_static_credential_type", content = "__field0")]
 #[serde(rename_all = "snake_case")]
 pub enum AzureStaticCredentials {
     AccessKey(String),
@@ -566,3 +587,87 @@ pub enum Credentials {
     Gcs(GcsCredentials),
     Azure(AzureCredentials),
 }
+
+#[cfg(test)]
+#[allow(clippy::panic, clippy::unwrap_used, clippy::expect_used)]
+mod tests {
+    use crate::{
+        ObjectStoreConfig, RepositoryConfig, config::S3Options,
+        strategies::repository_config, virtual_chunks::VirtualChunkContainer,
+    };
+
+    use proptest::prelude::*;
+
+    #[icechunk_macros::test]
+    fn test_config_serialization() {
+        let mut config = RepositoryConfig::default();
+        let bytes = rmp_serde::to_vec(&config).unwrap();
+        let roundtrip = rmp_serde::from_slice(&bytes).unwrap();
+        assert_eq!(config, roundtrip);
+
+        config.set_virtual_chunk_container(
+            VirtualChunkContainer::new(
+                "s3://foo/bar/".to_string(),
+                ObjectStoreConfig::S3(S3Options {
+                    region: Some("us-east-1".to_string()),
+                    endpoint_url: None,
+                    anonymous: false,
+                    allow_http: false,
+                    force_path_style: false,
+                    network_stream_timeout_seconds: None,
+                }),
+            )
+            .unwrap(),
+        );
+        let bytes = rmp_serde::to_vec(&config).unwrap();
+        let roundtrip = rmp_serde::from_slice(&bytes).unwrap();
+        assert_eq!(config, roundtrip);
+    }
+
+    #[icechunk_macros::test]
+    fn test_virtual_container_serialization() {
+        let container = VirtualChunkContainer::new(
+            "s3://foo/bar/".to_string(),
+            ObjectStoreConfig::S3(S3Options {
+                region: Some("us-east-1".to_string()),
+                endpoint_url: None,
+                anonymous: false,
+                allow_http: false,
+                force_path_style: false,
+                network_stream_timeout_seconds: None,
+            }),
+        )
+        .unwrap();
+        let bytes = rmp_serde::to_vec(&container).unwrap();
+        let roundtrip = rmp_serde::from_slice(&bytes).unwrap();
+        assert_eq!(container, roundtrip);
+    }
+
+    #[icechunk_macros::test]
+    fn test_object_store_config_serialization() {
+        let config = ObjectStoreConfig::S3(S3Options {
+            region: Some("us-east-1".to_string()),
+            endpoint_url: None,
+            anonymous: false,
+            allow_http: false,
+            force_path_style: false,
+            network_stream_timeout_seconds: None,
+        });
+        let bytes = rmp_serde::to_vec(&config).unwrap();
+        let roundtrip = rmp_serde::from_slice(&bytes).unwrap();
+        assert_eq!(config, roundtrip);
+    }
+
+    proptest! {
+        #![proptest_config(ProptestConfig {
+            cases: 50, .. ProptestConfig::default()
+        })]
+
+        #[icechunk_macros::test]
+        fn test_config_roundtrip(config in repository_config() ) {
+            let bytes = rmp_serde::to_vec(&config).unwrap();
+            let roundtrip = rmp_serde::from_slice(&bytes).unwrap();
+            assert_eq!(config, roundtrip);
+        }
+    }
+}
