@@ -13,10 +13,11 @@ use futures::{
     Stream, StreamExt, TryStreamExt,
     stream::{FuturesOrdered, FuturesUnordered},
 };
+use itertools::Itertools;
 use regex::bytes::Regex;
 use serde::{Deserialize, Serialize};
 use thiserror::Error;
-use tokio::{task::JoinError, try_join};
+use tokio::{sync::AcquireError, task::JoinError, try_join};
 use tracing::{Instrument, debug, error, instrument, trace};
 
 use crate::{
@@ -28,7 +29,8 @@ use crate::{
         IcechunkFormatError, IcechunkFormatErrorKind, ManifestId, NodeId, Path,
         SnapshotId,
         snapshot::{
-            ManifestFileInfo, NodeData, Snapshot, SnapshotInfo, SnapshotProperties,
+            ManifestFileInfo, NodeData, NodeType, Snapshot, SnapshotInfo,
+            SnapshotProperties,
         },
         transaction_log::{Diff, DiffBuilder},
     },
@@ -50,6 +52,28 @@ pub enum VersionInfo {
     AsOf { branch: String, at: DateTime<Utc> },
 }
 
+#[derive(Debug, Clone, PartialEq, Eq)]
+pub enum RefVersionInfo {
+    SnapshotId(SnapshotId),
+    TagRef(String),
+    BranchTipRef(String),
+}
+
+impl TryFrom<&VersionInfo> for RefVersionInfo {
+    type Error = (String, DateTime<Utc>);
+
+    fn try_from(value: &VersionInfo) -> Result<Self, Self::Error> {
+        match value {
+            VersionInfo::SnapshotId(id) => Ok(RefVersionInfo::SnapshotId(id.clone())),
+            VersionInfo::TagRef(name) => Ok(RefVersionInfo::TagRef(name.clone())),
+            VersionInfo::BranchTipRef(name) => {
+                Ok(RefVersionInfo::BranchTipRef(name.clone()))
+            }
+            VersionInfo::AsOf { branch, at } => Err((branch.clone(), *at)),
+        }
+    }
+}
+
 #[derive(Debug, Error)]
 #[non_exhaustive]
 pub enum RepositoryErrorKind {
@@ -89,6 +113,8 @@ pub enum RepositoryErrorKind {
     IOError(#[from] std::io::Error),
     #[error("a concurrent task failed")]
     ConcurrencyError(#[from] JoinError),
+    #[error("the http request semaphore cannot be acquired")]
+    AcquireError(#[from] AcquireError),
     #[error("main branch cannot be deleted")]
     CannotDeleteMain,
     #[error("the storage used by this Icechunk repository is read-only: {0}")]
@@ -173,12 +199,14 @@ impl Repository {
             return Err(RepositoryErrorKind::ParentDirectoryNotClean.into());
         }
 
+        let max_concurrent_reqs = config.max_concurrent_requests();
         let create_branch = async move {
             // TODO: we could cache this first snapshot
             let asset_manager = AssetManager::new_no_cache(
                 Arc::clone(&storage_c),
                 storage_settings.clone(),
                 compression,
+                max_concurrent_reqs,
             );
             // On create we need to create the default branch
             let new_snapshot = Arc::new(Snapshot::initial()?);
@@ -298,6 +326,7 @@ impl Repository {
             storage_settings.clone(),
             config.caching(),
             config.compression().level(),
+            config.max_concurrent_requests(),
         ));
         Ok(Self {
             config,
@@ -426,6 +455,10 @@ impl Repository {
         &self.asset_manager
     }
 
+    pub fn authorized_virtual_container_prefixes(&self) -> HashSet<String> {
+        self.authorized_virtual_containers.keys().cloned().collect()
+    }
+
     /// Returns the sequence of parents of the current session, in order of latest first.
     #[instrument(skip(self))]
     pub async fn snapshot_ancestry(
@@ -467,6 +500,16 @@ impl Repository {
         self.snapshot_ancestry_arc(&snapshot_id).await
     }
 
+    #[instrument(skip(self))]
+    async fn ancestry_ref<'a>(
+        &'a self,
+        version: &RefVersionInfo,
+    ) -> RepositoryResult<impl Stream<Item = RepositoryResult<SnapshotInfo>> + 'a + use<'a>>
+    {
+        let snapshot_id = self.resolve_ref_version(version).await?;
+        self.snapshot_ancestry(&snapshot_id).await
+    }
+
     /// Create a new branch in the repository at the given snapshot id
     #[instrument(skip(self))]
     pub async fn create_branch(
@@ -535,7 +578,8 @@ impl Repository {
     pub async fn reset_branch(
         &self,
         branch: &str,
-        snapshot_id: &SnapshotId,
+        to_snapshot_id: &SnapshotId,
+        from_snapshot_id: Option<&SnapshotId>,
     ) -> RepositoryResult<()> {
         if !self.storage.can_write() {
             return Err(RepositoryErrorKind::ReadonlyStorage(
@@ -546,16 +590,19 @@ impl Repository {
         raise_if_invalid_snapshot_id(
             self.storage.as_ref(),
             &self.storage_settings,
-            snapshot_id,
+            to_snapshot_id,
         )
         .await?;
-        let branch_tip = self.lookup_branch(branch).await?;
+        let branch_tip = match from_snapshot_id {
+            Some(snap) => snap,
+            None => &self.lookup_branch(branch).await?,
+        };
         update_branch(
             self.storage.as_ref(),
             &self.storage_settings,
             branch,
-            snapshot_id.clone(),
-            Some(&branch_tip),
+            to_snapshot_id.clone(),
+            Some(branch_tip),
         )
         .await
         .err_into()
@@ -639,12 +686,12 @@ impl Repository {
     }
 
     #[instrument(skip(self))]
-    pub async fn resolve_version(
+    pub async fn resolve_ref_version(
         &self,
-        version: &VersionInfo,
+        version: &RefVersionInfo,
     ) -> RepositoryResult<SnapshotId> {
         match version {
-            VersionInfo::SnapshotId(sid) => {
+            RefVersionInfo::SnapshotId(sid) => {
                 raise_if_invalid_snapshot_id(
                     self.storage.as_ref(),
                     &self.storage_settings,
@@ -653,12 +700,12 @@ impl Repository {
                 .await?;
                 Ok(sid.clone())
             }
-            VersionInfo::TagRef(tag) => {
+            RefVersionInfo::TagRef(tag) => {
                 let ref_data =
                     fetch_tag(self.storage.as_ref(), &self.storage_settings, tag).await?;
                 Ok(ref_data.snapshot)
             }
-            VersionInfo::BranchTipRef(branch) => {
+            RefVersionInfo::BranchTipRef(branch) => {
                 let ref_data = fetch_branch_tip(
                     self.storage.as_ref(),
                     &self.storage_settings,
@@ -667,12 +714,22 @@ impl Repository {
                 .await?;
                 Ok(ref_data.snapshot)
             }
-            VersionInfo::AsOf { branch, at } => {
-                let tip = VersionInfo::BranchTipRef(branch.clone());
+        }
+    }
+
+    #[instrument(skip(self))]
+    pub async fn resolve_version(
+        &self,
+        version: &VersionInfo,
+    ) -> RepositoryResult<SnapshotId> {
+        match version.try_into() {
+            Ok(ref_version_info) => self.resolve_ref_version(&ref_version_info).await,
+            Err((branch, at)) => {
+                let tip = RefVersionInfo::BranchTipRef(branch.clone());
                 let snap = self
-                    .ancestry(&tip)
+                    .ancestry_ref(&tip)
                     .await?
-                    .try_skip_while(|parent| ready(Ok(&parent.flushed_at > at)))
+                    .try_skip_while(|parent| ready(Ok(parent.flushed_at > at)))
                     .take(1)
                     .try_collect::<Vec<_>>()
                     .await?;
@@ -680,7 +737,7 @@ impl Repository {
                     Some(snap) => Ok(snap.id),
                     None => Err(RepositoryErrorKind::InvalidAsOfSpec {
                         branch: branch.clone(),
-                        at: *at,
+                        at,
                     }
                     .into()),
                 }
@@ -701,8 +758,9 @@ impl Repository {
         to: &VersionInfo,
     ) -> SessionResult<Diff> {
         let from = self.resolve_version(from).await?;
+        let to = self.resolve_version(to).await?;
         let all_snaps = self
-            .ancestry(to)
+            .ancestry_ref(&RefVersionInfo::SnapshotId(to))
             .await?
             .try_take_while(|snap_info| ready(Ok(snap_info.id != from)))
             .try_collect::<Vec<_>>()
@@ -809,7 +867,12 @@ impl Repository {
             // TODO: unnest this code
             if let Ok(snap) = asset_manager.fetch_snapshot(&snapshot_id).await {
                 let snap_c = Arc::clone(&snap);
-                for node in snap.iter_arc() {
+                for node in snap
+                    .iter_arc(&Path::root())
+                    .filter_ok(|node| node.node_type() == NodeType::Array)
+                    // TODO: make configurable
+                    .take(50)
+                {
                     match node {
                         Err(err) => {
                             error!(error=%err, "Error retrieving snapshot nodes");
@@ -822,19 +885,18 @@ impl Repository {
                                         let manifest_id = manifest.object_id;
                                         if let Some(manifest_info) =
                                             snap_c.manifest_info(&manifest_id)
-                                        {
-                                            if loaded_refs + manifest_info.num_chunk_refs
+                                            && loaded_refs + manifest_info.num_chunk_refs
                                                 <= preload_config.max_total_refs()
-                                                && preload_config
-                                                    .preload_if()
-                                                    .matches(&node.path, &manifest_info)
-                                            {
-                                                let size_bytes = manifest_info.size_bytes;
-                                                let asset_manager =
-                                                    Arc::clone(&asset_manager);
-                                                let manifest_id_c = manifest_id.clone();
-                                                let path = node.path.clone();
-                                                futures.push(async move {
+                                            && preload_config
+                                                .preload_if()
+                                                .matches(&node.path, &manifest_info)
+                                        {
+                                            let size_bytes = manifest_info.size_bytes;
+                                            let asset_manager =
+                                                Arc::clone(&asset_manager);
+                                            let manifest_id_c = manifest_id.clone();
+                                            let path = node.path.clone();
+                                            futures.push(async move {
                                                     trace!("Preloading manifest {} for array {}", &manifest_id_c, path);
                                                     if let Err(err) = asset_manager
                                                         .fetch_manifest(
@@ -849,10 +911,8 @@ impl Repository {
                                                         );
                                                     }
                                                 });
-                                                loaded_manifests.insert(manifest_id);
-                                                loaded_refs +=
-                                                    manifest_info.num_chunk_refs;
-                                            }
+                                            loaded_manifests.insert(manifest_id);
+                                            loaded_refs += manifest_info.num_chunk_refs;
                                         }
                                     }
                                 }
@@ -902,13 +962,13 @@ fn validate_credentials(
     creds: &HashMap<String, Option<Credentials>>,
 ) -> RepositoryResult<()> {
     for (url_prefix, cred) in creds {
-        if let Some(cont) = config.get_virtual_chunk_container(url_prefix) {
-            if let Err(error) = cont.validate_credentials(cred.as_ref()) {
-                return Err(RepositoryErrorKind::StorageError(StorageErrorKind::Other(
-                    error,
-                ))
-                .into());
-            }
+        if let Some(cont) = config.get_virtual_chunk_container(url_prefix)
+            && let Err(error) = cont.validate_credentials(cred.as_ref())
+        {
+            return Err(RepositoryErrorKind::StorageError(StorageErrorKind::Other(
+                error,
+            ))
+            .into());
         }
     }
     Ok(())
