@@ -23,7 +23,8 @@ use aws_credential_types::provider::error::CredentialsError;
 use aws_sdk_s3::{
     Client,
     config::{
-        Builder, ConfigBag, Intercept, ProvideCredentials, Region, RuntimeComponents,
+        Builder, ConfigBag, IdentityCache, Intercept, ProvideCredentials, Region,
+        RuntimeComponents, StalledStreamProtectionConfig,
         interceptors::BeforeTransmitInterceptorContextMut,
     },
     error::{BoxError, SdkError},
@@ -40,7 +41,7 @@ use futures::{
 };
 use serde::{Deserialize, Serialize};
 use tokio::{io::AsyncRead, sync::OnceCell};
-use tracing::instrument;
+use tracing::{error, instrument};
 
 use super::{
     CHUNK_PREFIX, CONFIG_PATH, DeleteObjectsResult, FetchConfigResult, GetRefResult,
@@ -136,7 +137,7 @@ pub async fn mk_client(
 
     #[allow(clippy::unwrap_used)]
     let app_name = AppName::new("icechunk").unwrap();
-    let mut aws_config = aws_config::defaults(BehaviorVersion::v2025_01_17())
+    let mut aws_config = aws_config::defaults(BehaviorVersion::v2025_08_07())
         .region(region)
         .app_name(app_name);
 
@@ -144,6 +145,17 @@ pub async fn mk_client(
         aws_config = aws_config.endpoint_url(endpoint)
     }
 
+    let stalled_stream = if config.network_stream_timeout_seconds == Some(0) {
+        StalledStreamProtectionConfig::disabled()
+    } else {
+        StalledStreamProtectionConfig::enabled()
+            .grace_period(std::time::Duration::from_secs(
+                config.network_stream_timeout_seconds.unwrap_or(60) as u64,
+            ))
+            .build()
+    };
+    aws_config = aws_config.stalled_stream_protection(stalled_stream);
+
     match credentials {
         S3Credentials::FromEnv => {}
         S3Credentials::Anonymous => aws_config = aws_config.no_credentials(),
@@ -171,10 +183,19 @@ pub async fn mk_client(
         .with_max_backoff(core::time::Duration::from_millis(
             settings.retries().max_backoff_ms() as u64,
         ));
+
     let mut s3_builder = Builder::from(&aws_config.load().await)
         .force_path_style(config.force_path_style)
         .retry_config(retry_config);
 
+    // credentials may take a while to refresh, defaults are too strict
+    let id_cache = IdentityCache::lazy()
+        .load_timeout(core::time::Duration::from_secs(120))
+        .buffer_time(core::time::Duration::from_secs(120))
+        .build();
+
+    s3_builder = s3_builder.identity_cache(id_cache);
+
     if !extra_read_headers.is_empty() || !extra_write_headers.is_empty() {
         s3_builder = s3_builder.interceptor(ExtraHeadersInterceptor {
             extra_read_headers,
@@ -304,10 +325,10 @@ impl S3Storage {
             .bucket(self.bucket.clone())
             .key(key);
 
-        if settings.unsafe_use_metadata() {
-            if let Some(ct) = content_type {
-                b = b.content_type(ct)
-            };
+        if settings.unsafe_use_metadata()
+            && let Some(ct) = content_type
+        {
+            b = b.content_type(ct);
         }
 
         if settings.unsafe_use_metadata() {
@@ -837,6 +858,8 @@ impl Storage for S3Storage {
                 let code = err.as_service_error().and_then(|e| e.code()).unwrap_or("");
                 if code.contains("PreconditionFailed")
                     || code.contains("ConditionalRequestConflict")
+                    // ConcurrentModification sent by Ceph Object Gateway
+                    || code.contains("ConcurrentModification")
                 {
                     Ok(WriteRefResult::WontOverwrite)
                 } else {
@@ -888,11 +911,11 @@ impl Storage for S3Storage {
         let mut sizes = HashMap::new();
         let mut ids = Vec::new();
         for (id, size) in batch.into_iter() {
-            if let Ok(key) = self.get_path_str(prefix, id.as_str()) {
-                if let Ok(ident) = ObjectIdentifier::builder().key(key.clone()).build() {
-                    ids.push(ident);
-                    sizes.insert(key, size);
-                }
+            if let Ok(key) = self.get_path_str(prefix, id.as_str())
+                && let Ok(ident) = ObjectIdentifier::builder().key(key.clone()).build()
+            {
+                ids.push(ident);
+                sizes.insert(key, size);
             }
         }
 
@@ -1017,7 +1040,12 @@ impl ProvideRefreshableCredentials {
     async fn provide(
         &self,
     ) -> Result<aws_credential_types::Credentials, CredentialsError> {
-        let creds = self.0.get().await.map_err(CredentialsError::not_loaded)?;
+        let creds = self
+            .0
+            .get()
+            .await
+            .inspect_err(|err| error!(error = err, "Cannot load credentials"))
+            .map_err(CredentialsError::not_loaded)?;
         let creds = aws_credential_types::Credentials::new(
             creds.access_key_id,
             creds.secret_access_key,
@@ -1060,6 +1088,7 @@ mod tests {
             allow_http: true,
             anonymous: false,
             force_path_style: false,
+            network_stream_timeout_seconds: None,
         };
         let credentials = S3Credentials::Static(S3StaticCredentials {
             access_key_id: "access_key_id".to_string(),
@@ -1082,7 +1111,7 @@ mod tests {
 
         assert_eq!(
             serialized,
-            r#"{"config":{"region":"us-west-2","endpoint_url":"http://localhost:9000","anonymous":false,"allow_http":true,"force_path_style":false},"credentials":{"s3_credential_type":"static","access_key_id":"access_key_id","secret_access_key":"secret_access_key","session_token":"session_token","expires_after":null},"bucket":"bucket","prefix":"prefix","can_write":true,"extra_read_headers":[],"extra_write_headers":[]}"#
+            r#"{"config":{"region":"us-west-2","endpoint_url":"http://localhost:9000","anonymous":false,"allow_http":true,"force_path_style":false,"network_stream_timeout_seconds":null},"credentials":{"s3_credential_type":"static","access_key_id":"access_key_id","secret_access_key":"secret_access_key","session_token":"session_token","expires_after":null},"bucket":"bucket","prefix":"prefix","can_write":true,"extra_read_headers":[],"extra_write_headers":[]}"#
         );
 
         let deserialized: S3Storage = serde_json::from_str(&serialized).unwrap();
@@ -1098,6 +1127,7 @@ mod tests {
                 allow_http: true,
                 anonymous: false,
                 force_path_style: false,
+                network_stream_timeout_seconds: None,
             },
             "bucket".to_string(),
             Some("prefix".to_string()),
