@@ -44,7 +44,9 @@ pub type ContainerName = String;
 pub struct VirtualChunkContainer {
     // name is no longer needed, but we keep it for compatibility with
     // old serialized configurations
-    #[serde(skip_serializing_if = "Option::is_none")]
+    // We use default because for a while after 1.0 we were skipping
+    // serialization if None. Then we learned rmp_serde doesn't really
+    // support optional fields, so serialization was broken
     #[serde(default)]
     pub name: Option<ContainerName>,
 
@@ -56,6 +58,12 @@ pub type VirtualChunkCredentialsError = String;
 
 impl VirtualChunkContainer {
     pub fn new(url_prefix: String, store: ObjectStoreConfig) -> Result<Self, String> {
+        if !url_prefix.ends_with('/') {
+            return Err(
+                "VirtualChunkContainer url_prefix must end in a / character".to_string()
+            );
+        }
+
         let url = Url::parse(url_prefix.as_str()).map_err(|e| e.to_string())?;
         match (url.scheme(), &store) {
             (
@@ -116,9 +124,9 @@ impl VirtualChunkContainer {
                     );
                 }
             }
-            (scheme, _) => {
+            (scheme, store) => {
                 return Err(format!(
-                    "Invalid url prefix scheme ({scheme}) for this object store config"
+                    "Invalid url prefix scheme ({scheme}) for this object store config: ({store:?})"
                 ));
             }
         };
@@ -249,13 +257,17 @@ impl VirtualChunkResolver {
         credentials: HashMap<String, Option<Credentials>>,
         settings: storage::Settings,
     ) -> Self {
+        fn add_trailing(s: String) -> String {
+            if s.ends_with('/') { s } else { format!("{s}/") }
+        }
+
         // we need to validate the containers because they can come from persisted config
         // they can be manipulated to have invalid url prefixes
         // We chose to silently filter out invalid containers
 
         let mut containers = containers
             .filter_map(|cont| {
-                VirtualChunkContainer::new(cont.url_prefix, cont.store)
+                VirtualChunkContainer::new(add_trailing(cont.url_prefix), cont.store)
                     .inspect_err(|err| {
                         tracing::warn!(
                             "Invalid virtual chunk container, ignoring it: {err}"
@@ -265,6 +277,8 @@ impl VirtualChunkResolver {
             })
             .collect::<Vec<_>>();
         sort_containers(&mut containers);
+        let credentials =
+            credentials.into_iter().map(|(k, v)| (add_trailing(k), v)).collect();
         VirtualChunkResolver { containers, credentials, settings, fetchers: new_cache() }
     }
 
@@ -279,7 +293,8 @@ impl VirtualChunkResolver {
         &self,
         chunk_location: &Url,
     ) -> Result<Arc<dyn ChunkFetcher>, VirtualReferenceError> {
-        let cont = find_container(chunk_location.to_string().as_str(), &self.containers)
+        let cont = self
+            .matching_container(chunk_location.to_string().as_str())
             .ok_or_else(|| {
                 VirtualReferenceErrorKind::NoContainerForUrl(chunk_location.to_string())
             })?;
@@ -303,8 +318,12 @@ impl VirtualChunkResolver {
         checksum: Option<&Checksum>,
     ) -> Result<Bytes, VirtualReferenceError> {
         // this resolves things like /../
-        let url = Url::parse(chunk_location)
-            .map_err(VirtualReferenceErrorKind::CannotParseUrl)?;
+        let url = Url::parse(chunk_location).map_err(|e| {
+            VirtualReferenceErrorKind::CannotParseUrl {
+                cause: e,
+                url: chunk_location.to_string(),
+            }
+        })?;
         let fetcher = self.get_fetcher(&url).await?;
         fetcher.fetch_chunk(&url, range, checksum).await
     }
@@ -360,7 +379,7 @@ impl VirtualChunkResolver {
                     opts
                 } else {
                     &S3Options {
-                        endpoint_url: Some("https://fly.storage.tigris.dev".to_string()),
+                        endpoint_url: Some("https://t3.storage.dev".to_string()),
                         ..opts.clone()
                     }
                 };
@@ -725,7 +744,8 @@ mod tests {
                     endpoint_url: None,
                     anonymous: false,
                     allow_http: false,
-                    force_path_style: false
+                    force_path_style: false,
+                    network_stream_timeout_seconds: None,
                 })
             )
             .is_err()
@@ -738,7 +758,8 @@ mod tests {
                     endpoint_url: None,
                     anonymous: false,
                     allow_http: false,
-                    force_path_style: false
+                    force_path_style: false,
+                    network_stream_timeout_seconds: None,
                 })
             )
             .is_err()
@@ -751,7 +772,8 @@ mod tests {
                     endpoint_url: None,
                     anonymous: false,
                     allow_http: false,
-                    force_path_style: false
+                    force_path_style: false,
+                    network_stream_timeout_seconds: None,
                 })
             )
             .is_err()
@@ -764,7 +786,8 @@ mod tests {
                     endpoint_url: None,
                     anonymous: false,
                     allow_http: false,
-                    force_path_style: false
+                    force_path_style: false,
+                    network_stream_timeout_seconds: None,
                 })
             )
             .is_err()
@@ -777,7 +800,8 @@ mod tests {
                     endpoint_url: None,
                     anonymous: false,
                     allow_http: false,
-                    force_path_style: false
+                    force_path_style: false,
+                    network_stream_timeout_seconds: None,
                 })
             )
             .is_err()
@@ -790,7 +814,8 @@ mod tests {
                     endpoint_url: None,
                     anonymous: false,
                     allow_http: false,
-                    force_path_style: false
+                    force_path_style: false,
+                    network_stream_timeout_seconds: None,
                 })
             )
             .is_err()
@@ -803,7 +828,8 @@ mod tests {
                     endpoint_url: None,
                     anonymous: false,
                     allow_http: false,
-                    force_path_style: false
+                    force_path_style: false,
+                    network_stream_timeout_seconds: None,
                 })
             )
             .is_err()
@@ -813,7 +839,7 @@ mod tests {
     #[tokio::test]
     async fn test_cannot_resolve_for_nonexistent_container() {
         let container = VirtualChunkContainer::new(
-            "file:///foo".to_string(),
+            "file:///foo/".to_string(),
             ObjectStoreConfig::LocalFileSystem("/example".into()),
         )
         .unwrap();
@@ -837,7 +863,7 @@ mod tests {
     #[tokio::test]
     async fn test_cannot_resolve_for_unauthorized_container() {
         let container = VirtualChunkContainer::new(
-            "file:///example".to_string(),
+            "file:///example/".to_string(),
             ObjectStoreConfig::LocalFileSystem("/example".into()),
         )
         .unwrap();
@@ -854,7 +880,7 @@ mod tests {
             Err(VirtualReferenceError {
                 kind: VirtualReferenceErrorKind::UnauthorizedVirtualChunkContainer(cont),
                 ..
-            }) if cont.url_prefix() == "file:///example"
+            }) if cont.url_prefix() == "file:///example/"
         ));
     }
 
