@@ -37,6 +37,7 @@ async fn mk_s3_storage(prefix: &str) -> StorageResult<Arc<dyn Storage + Send + S
             allow_http: true,
             anonymous: false,
             force_path_style: true,
+            network_stream_timeout_seconds: None,
         },
         "testbucket".to_string(),
         Some(prefix.to_string()),
@@ -72,6 +73,7 @@ async fn mk_s3_object_store_storage(
                 allow_http: true,
                 anonymous: false,
                 force_path_style: true,
+                network_stream_timeout_seconds: None,
             }),
         )
         .await?,
@@ -656,7 +658,7 @@ pub async fn test_storage_classes() -> Result<(), Box<dyn std::error::Error>> {
     let out = client
         .list_objects_v2()
         .bucket(common::get_aws_integration_bucket()?)
-        .prefix(format!("{prefix}/chunks"))
+        .prefix(format!("{prefix}/manifests"))
         .into_paginator()
         .send()
         .collect::<Vec<_>>()
