@@ -50,6 +50,7 @@ fn minio_s3_config() -> (S3Options, S3Credentials) {
         allow_http: true,
         anonymous: false,
         force_path_style: true,
+        network_stream_timeout_seconds: None,
     };
     let credentials = S3Credentials::Static(S3StaticCredentials {
         access_key_id: "minio123".into(),
@@ -112,34 +113,36 @@ async fn create_local_repository(
 
     let mut containers = vec![
         VirtualChunkContainer::new(
-            "s3://testbucket".to_string(),
+            "s3://testbucket/".to_string(),
             ObjectStoreConfig::S3(S3Options {
                 region: Some("us-east-1".to_string()),
                 endpoint_url: None,
                 anonymous: true,
                 allow_http: false,
                 force_path_style: false,
+                network_stream_timeout_seconds: None,
             }),
         )
         .unwrap(),
         VirtualChunkContainer::new(
-            "s3://earthmover-sample-data".to_string(),
+            "s3://earthmover-sample-data/".to_string(),
             ObjectStoreConfig::S3(S3Options {
                 region: Some("us-east-1".to_string()),
                 endpoint_url: None,
                 anonymous: true,
                 allow_http: false,
                 force_path_style: false,
+                network_stream_timeout_seconds: None,
             }),
         )
         .unwrap(),
         VirtualChunkContainer::new(
-            "gcs://testbucket".to_string(),
+            "gcs://testbucket/".to_string(),
             ObjectStoreConfig::Gcs(Default::default()),
         )
         .unwrap(),
         VirtualChunkContainer::new(
-            "gcs://earthmover-sample-data".to_string(),
+            "gcs://earthmover-sample-data/".to_string(),
             ObjectStoreConfig::Gcs(Default::default()),
         )
         .unwrap(),
@@ -160,7 +163,7 @@ async fn create_local_repository(
     .into();
 
     if let Some(chunks_path) = chunks_path {
-        let prefix = format!("file://{}", chunks_path.to_str().unwrap());
+        let prefix = format!("file://{}/", chunks_path.to_str().unwrap());
         containers.push(
             VirtualChunkContainer::new(
                 prefix.clone(),
@@ -183,20 +186,21 @@ async fn create_minio_repository() -> Repository {
 
     let containers = vec![
         VirtualChunkContainer::new(
-            "s3://testbucket".to_string(),
+            "s3://testbucket/".to_string(),
             ObjectStoreConfig::S3Compatible(S3Options {
                 region: Some(String::from("us-east-1")),
                 endpoint_url: Some("http://localhost:9000".to_string()),
                 anonymous: false,
                 allow_http: true,
                 force_path_style: true,
+                network_stream_timeout_seconds: None,
             }),
         )
         .unwrap(),
     ];
 
     let credentials = [(
-        "s3://testbucket".to_string(),
+        "s3://testbucket/".to_string(),
         Some(Credentials::S3(S3Credentials::Static(S3StaticCredentials {
             access_key_id: "minio123".to_string(),
             secret_access_key: "minio123".to_string(),
@@ -540,7 +544,7 @@ async fn test_zarr_store_virtual_refs_minio_set_and_get()
     };
     assert!(matches!(
                 store.set_virtual_ref("array/c/0/0/0", bad_ref, true).await,
-                Err(StoreError{kind: StoreErrorKind::InvalidVirtualChunkContainer { chunk_location },..}) if chunk_location == bad_location.0));
+                Err(StoreError{kind: StoreErrorKind::InvalidVirtualChunkContainer { chunk_location },..}) if chunk_location == bad_location.url()));
     Ok(())
 }
 
@@ -688,29 +692,31 @@ async fn test_zarr_store_with_multiple_virtual_chunk_containers()
 
     let containers = vec![
         VirtualChunkContainer::new(
-            "s3://testbucket".to_string(),
+            "s3://testbucket/".to_string(),
             ObjectStoreConfig::S3Compatible(S3Options {
                 region: Some(String::from("us-east-1")),
                 endpoint_url: Some("http://localhost:9000".to_string()),
                 anonymous: false,
                 allow_http: true,
                 force_path_style: true,
+                network_stream_timeout_seconds: None,
             }),
         )
         .unwrap(),
         VirtualChunkContainer::new(
-            format!("file://{}", chunk_dir.path().to_str().unwrap()),
+            format!("file://{}/", chunk_dir.path().to_str().unwrap()),
             ObjectStoreConfig::LocalFileSystem(PathBuf::new()),
         )
         .unwrap(),
         VirtualChunkContainer::new(
-            "s3://earthmover-sample-data".to_string(),
+            "s3://earthmover-sample-data/".to_string(),
             ObjectStoreConfig::S3(S3Options {
                 region: Some(String::from("us-east-1")),
                 endpoint_url: None,
                 anonymous: true,
                 allow_http: false,
                 force_path_style: false,
+                network_stream_timeout_seconds: None,
             }),
         )
         .unwrap(),
